// ImGUI_Test1.cpp : Defines the entry point for the application.
//
 #pragma comment(linker, "/SUBSYSTEM:windows /ENTRY:mainCRTStartup") // hides the debug console when running the .exe

#include "UseImGUI.h"
#include "SerialFunctions.h"
#include "glad/glad.h"
#include "GLFW/glfw3.h"

using namespace std;

const int width = 1280;
const int height = 720;

int main() {

	// This Tell the ImGui code to not save settings in an .ini file
	ImGui::CreateContext();
	ImGuiIO& io = ImGui::GetIO();
	io.IniFilename = NULL;

	// Setup window
	if (!glfwInit())
		return 1;

	// GL 3.0 + GLSL 130
	const char* glsl_version = "#version 130";
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);

	// Code to recognize High DPI monitors
	glfwWindowHint(GLFW_SCALE_TO_MONITOR, GLFW_TRUE);

	// Create window with graphics context
	GLFWwindow* window = glfwCreateWindow(width, height, "Ricardo's Serial Monitor - FabAcademy", NULL, NULL);

	// Lock the GLFW window dimensions
	glfwSetWindowSizeLimits(window, width, height, width, height);
	
	// Disables resizing altogether
	glfwSetWindowAttrib(window, GLFW_RESIZABLE, GLFW_FALSE); 
	

	if (window == NULL)
		return 1;
	glfwMakeContextCurrent(window);
	glfwSwapInterval(1); // Enable vsync

	if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) // Ties OpenGL to program window
		throw("unable to context to OpenGL");

	int screen_width, screen_height;
	glfwGetFramebufferSize(window, &screen_width, &screen_height);
	glViewport(0, 0, screen_width, screen_height);

	UseImGUI myimgui;
	// Set the scaling factor to 2 (double the size)
	ImGui::GetIO().FontGlobalScale = 2.0f;

	myimgui.Init(window, glsl_version);

	while (!glfwWindowShouldClose(window)) {
		glfwPollEvents();
		glClearColor(0.45f, 0.55f, 0.60f, 1.00F); // f means float
		glClear(GL_COLOR_BUFFER_BIT);
		myimgui.NewFrame();
		myimgui.Update();
		// ImGui::ShowDemoWindow();
		myimgui.Render();

		glfwSwapBuffers(window);

	}
	myimgui.Shutdown();
	return 0;
}
