// all these functions where given to my by ChatGPT 4.0

#include "SerialFunctions.h"
#include "windows.h"
#include "Winerror.h"
#include "iostream"
#include "vector"
#include "string"
#include "Setupapi.h"
#include "devguid.h"
#include "cfgmgr32.h" 




DWORD useSerial::bytesRead = 0; 

char useSerial::inputBuffer[1024 * 1] = "";
bool useSerial::inputEntered = false;



std::vector<std::string> useSerial::getAvailableSerialPorts() {
    // Define a vector to store the available COM ports
    std::vector<std::string> com_ports_list;

    // Use the SetupDiGetClassDevs() function to retrieve a list of all connected serial devices
    HDEVINFO device_info_set = SetupDiGetClassDevs(&GUID_DEVCLASS_PORTS, NULL, NULL, DIGCF_PRESENT);
    if (device_info_set == INVALID_HANDLE_VALUE) {
        std::cerr << "Failed to get device info set: " << GetLastError() << std::endl;
        return com_ports_list;
    }

    // Define a SP_DEVINFO_DATA struct to store device info
    SP_DEVINFO_DATA device_info_data;
    device_info_data.cbSize = sizeof(SP_DEVINFO_DATA);

    // Loop through each connected device and check if it's a serial device
    for (DWORD i = 0; SetupDiEnumDeviceInfo(device_info_set, i, &device_info_data); i++) {
        // Define a buffer to store the friendly name of the device
        TCHAR friendly_name[MAX_PATH];
        DWORD friendly_name_size = MAX_PATH;

        // Retrieve the friendly name of the device
        if (SetupDiGetDeviceRegistryProperty(device_info_set, &device_info_data, SPDRP_FRIENDLYNAME, NULL, (PBYTE)friendly_name, friendly_name_size, NULL)) {
            // Check if the device is a serial device
            if (std::string(friendly_name).find("COM") != std::string::npos) {
                // Add the available COM port to the vector
                std::cout << "Adding device with friendly name: " << friendly_name << std::endl;
                com_ports_list.push_back(std::string(friendly_name));
            }
            else {
                std::cout << "Device with friendly name: " << friendly_name << " is not a serial device" << std::endl;
            }
        }
        else {
            std::cerr << "Failed to get friendly name: " << GetLastError() << std::endl;
        }
    }

    // Cleanup the device info set
    SetupDiDestroyDeviceInfoList(device_info_set);

    // Return the vector of available COM ports
    for (const auto& port : com_ports_list) {
        std::cout << "Available port: " << port << std::endl;
    }
    std::cout << "com_ports_list size: " << com_ports_list.size() << std::endl;
    return com_ports_list;
}

bool useSerial::isPortAvailable(std::string portName) {
    HANDLE hComm = CreateFile(portName.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (hComm == INVALID_HANDLE_VALUE) {
        // Port is not available, it's already in use
        std::cerr << "useSerial:: Selected COM port is already in use" << std::endl;
        return false;
    }
    else {
        // Port is available, close the handle and return true
        // std::cerr << "useSerial:: Selected COM port is Available " << std::endl;
        CloseHandle(hComm);
        return true;
    }
}

useSerial::SerialError useSerial::connectSerialPort(std::string portName, int baudRate) {
    useSerial::f_name = portName;
    // std::cout << "useSerial:: Parameter input Data bits: " << +dataBits << std::endl; // small number in equivalent to CHAR datatypes are reserved characters on the print command, they must be explicitly noted as numerical values with static_cast or unary plus
    // std::cout << "useSerial:: Parameter input Parity: " << +parity << std::endl; // small number in equivalent to CHAR datatypes are reserved characters on the print command, they must be explicitly noted as numerical values with static_cast or unary plus
    // std::cout << "useSerial:: Parameter input Stop bits: " << +stopBits << std::endl; // small number in equivalent to CHAR datatypes are reserved characters on the print command, they must be explicitly noted as numerical values with static_cast or unary plus
    // std::cout << "useSerial:: Parameter input readTimeout: " << readTimeout << std::endl;
    // std::cout << "useSerial:: Parameter input writeTimeout bits: " << writeTimeout << std::endl;
    // std::cerr << "useSerial:: 1 COM port connection status " << useSerial::connectedCOM << std::endl;

    // it will only try to connect if the COM port is available and it's we're not already connected
    if (useSerial::isPortAvailable(portName) && useSerial::connectedCOM != true) {
        // Open the serial port
        hSerial = CreateFile(portName.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);

        if (hSerial == INVALID_HANDLE_VALUE) {

            DWORD lastError = GetLastError();
            LPSTR messageBuffer = nullptr;
            size_t size = FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, lastError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR)&messageBuffer, 0, NULL);
            std::cout << "userSerial::Failed to open serial port with error code " << lastError << ": " << messageBuffer << std::endl;
            LocalFree(messageBuffer);

            if (lastError == ERROR_FILE_NOT_FOUND) {
                std::cout << "userSerial::Failed to create handle to serial port: " << GetLastError() << std::endl;
                return SERIAL_ERROR_PORT_NOT_FOUND;
            }
            else if (lastError == ERROR_ACCESS_DENIED) {
                return SERIAL_ERROR_PERMISSION_DENIED;
            }
            else if (lastError == ERROR_ALREADY_EXISTS) {
                return SERIAL_ERROR_PORT_IN_USE;
            }
            else {
                return SERIAL_ERROR_FAILED_TO_OPEN_PORT;
            }
        }

        if (hSerial == INVALID_HANDLE_VALUE) {
            DWORD lastError = GetLastError();
            if (lastError == ERROR_FILE_NOT_FOUND) {
                return SERIAL_ERROR_PORT_NOT_FOUND;
            }
            else if (lastError == ERROR_ACCESS_DENIED) {
                return SERIAL_ERROR_PERMISSION_DENIED;
            }
            else if (lastError == ERROR_ALREADY_EXISTS) {
                return SERIAL_ERROR_PORT_IN_USE;
            }
            else {
                return SERIAL_ERROR_FAILED_TO_OPEN_PORT;
            }
        }

        // Set the serial port parameters
        DCB dcbSerialParams = { 0 };
        dcbSerialParams.DCBlength = sizeof(dcbSerialParams);

        if (!GetCommState(hSerial, &dcbSerialParams)) {
            std::cerr << "useSerial:: Failed to get serial port parameters for port " << portName << std::endl;
            return SERIAL_ERROR_FAILED_TO_GET_STATE;
        }

        dcbSerialParams.BaudRate = baudRate;
        dcbSerialParams.ByteSize = useSerial::dataBits;
        dcbSerialParams.StopBits = useSerial::stopBits;
        dcbSerialParams.Parity = useSerial::parity;

        if (!SetCommState(hSerial, &dcbSerialParams)) {

            // Failed to set serial port parameters
            std::cerr << "useSerial:: Failed to set serial port parameters for port " << portName << std::endl;
            CloseHandle(hSerial);
            return SERIAL_ERROR_FAILED_TO_SET_STATE;
        }
        else {
            // Set read and write timeouts

            COMMTIMEOUTS timeouts = { 0 };
            timeouts.ReadIntervalTimeout = readTimeout;
            timeouts.ReadTotalTimeoutConstant = readTimeout;
            timeouts.ReadTotalTimeoutMultiplier = 0;
            timeouts.WriteTotalTimeoutConstant = writeTimeout;
            timeouts.WriteTotalTimeoutMultiplier = 0;
            if (!SetCommTimeouts(hSerial, &timeouts)) {

                // Failed to set serial port timeouts
                std::cerr << "useSerial:: Failed to set serial port timeouts for port " << portName << std::endl;
                CloseHandle(hSerial);
                return SERIAL_ERROR_FAILED_TO_SET_STATE;
            }

            // std::cout << "useSerial:: Data bits: " << +dataBits << std::endl; // small number in equivalent to CHAR datatypes are reserved characters, they must be explicitly noted as numerical values with static_cast or unary plus
            // std::cout << "useSerial:: Parity: " << +parity << std::endl; // small number in equivalent to CHAR datatypes are reserved characters, they must be explicitly noted as numerical values with static_cast or unary plus
            // std::cout << "useSerial:: Stop bits: " << +stopBits << std::endl; // small number in equivalent to CHAR datatypes are reserved characters, they must be explicitly noted as numerical values with static_cast or unary plus
            // std::cout << "useSerial:: readTimeout: " << readTimeout << std::endl;
            // std::cout << "useSerial:: writeTimeout: " << writeTimeout << std::endl;

            // Return the error code
            // std::cerr << "useSerial::2 COM port connection status " << useSerial::connectedCOM << std::endl;
            useSerial::connectedCOM = true;
            // std::cerr << "useSerial::2 COM port connection status " << useSerial::connectedCOM << std::endl;
            std::cerr << "useSerial:: Successfully connected to " << portName << std::endl;
            
            PurgeComm(hSerial, PURGE_RXCLEAR | PURGE_TXCLEAR);
            Sleep(1000);
        }

    }
    return SERIAL_ERROR_SUCCESS;
}

void useSerial::closeSerialPort(int port) {
    // std::cout << "useSerial::closeSerialPort Start" << std::endl;
    PurgeComm(hSerial, PURGE_RXCLEAR | PURGE_TXCLEAR);
    Sleep(100);
    // std::cerr << "useSerial:: COM port connection status " << useSerial::connectedCOM << std::endl;
    if (useSerial::connectedCOM == true) {
        if (hSerial != INVALID_HANDLE_VALUE) {
            DWORD error = GetLastError();
            // std::cout << "Error after (hSerial != INVALID_HANDLE_VALUE): " << error << std::endl;
            if (CloseHandle(hSerial)) {
                DWORD error = GetLastError();
                // std::cout << "Error after CloseHandle(hSerial): " << error << std::endl;

                std::cout << "useSerial:: Successfully disconnected from " << useSerial::f_name << std::endl;
                useSerial::connectedCOM = false;
            }
            else {
                std::cout << "Error closing serial port: " << GetLastError() << std::endl;
            }
            hSerial = INVALID_HANDLE_VALUE;
        }
    } else {
        std::cout << "useSerial:: Can't disconnect if not connected" << std::endl;
    }
    // std::cout << "useSerial::closeSerialPort End" << std::endl;
}

bool useSerial::sendSerialData(const char* buffer, int SendbufferSize) {

    DWORD bytesWritten;

    if (connectedCOM) {
        if (!WriteFile(hSerial, buffer, SendbufferSize, &bytesWritten, NULL)) {
            std::cerr << "useSerial:: Failed to write to serial port" << std::endl;
            return false;
        }
        return true;
    }
}

bool useSerial::readSerialData(char* buffer, int readBufferSize) {
    COMSTAT status = { 0 };
    unsigned int toRead = 0;

    if (connectedCOM) {
        if (!ClearCommError(hSerial, NULL, &status)) {
            DWORD errorCode = GetLastError();
            std::cerr << "useSerial:: Failed to clear comm error. Error code: " << errorCode << std::endl;
            Sleep(500);
            return false;
        }


        if (status.cbInQue > 0) {
            if (status.cbInQue > readBufferSize) {
                toRead = useSerial::readBufferSize;
                readBufferSize = 0;
                // std::cout << "userSerial::toRead was set to " << toRead << " After if (status.cbInQue > bufferSize)" << std::endl;
                // Sleep(500);
            }
            else {
                toRead = status.cbInQue;
                // std::cout << "userSerial::toRead was set to " << toRead << " After else" << std::endl;
                // Sleep(500);
            }

            if (!ReadFile(hSerial, buffer, toRead, &useSerial::bytesRead, NULL)) {
                DWORD errorCode = GetLastError();
                std::cerr << "useSerialData:: Failed to read from serial port. Error code: " << errorCode << std::endl;
                return false;
            }

            buffer[useSerial::bytesRead] = '\0'; // null-terminate received data
            // std::cout << "userSerial::Read " << useSerial::bytesRead << " bytes from serial port." << std::endl;
            // std::cout << "userSerial::toRead was set to " << toRead << " After else" << std::endl;
            // Sleep(500);
        }
    }

    return true;
}

bool useSerial::isConfigured() {
    return hSerial != INVALID_HANDLE_VALUE;
}