#ifndef USE_SERIAL_H
#define USE_SERIAL_H

#pragma once

#include "vector"
#include "string"
#include "windows.h"

class useSerial {
	public:
		enum SerialError {
			SERIAL_ERROR_SUCCESS = 0,
			SERIAL_ERROR_PORT_NOT_FOUND = 1,
			SERIAL_ERROR_PORT_IN_USE = 2,
			SERIAL_ERROR_PERMISSION_DENIED = 3,
			SERIAL_ERROR_INVALID_BAUD_RATE = 4,
			SERIAL_ERROR_INVALID_DATA_BITS = 5,
			SERIAL_ERROR_INVALID_STOP_BITS = 6,
			SERIAL_ERROR_INVALID_PARITY = 7,
			SERIAL_ERROR_FAILED_TO_GET_STATE = 8,
			SERIAL_ERROR_FAILED_TO_SET_STATE = 9,
			SERIAL_ERROR_TIMEOUT = 10,
			SERIAL_ERROR_FAILED_TO_OPEN_PORT = 11
		};

		std::vector<std::string> getAvailableSerialPorts();
		bool useSerial::isPortAvailable(std::string portName);
		SerialError connectSerialPort(std::string portName, int baudRate);
		void closeSerialPort(int port);
		bool sendSerialData(const char* buffer, int SendbufferSize);
		bool useSerial::readSerialData(char* buffer, int ReadbufferSize = useSerial::readBufferSize);
		bool useSerial::isConfigured();


		static DWORD bytesRead;
		// Defines the variable with the size of the buffer of data read from serial
		static const int readBufferSize = 256;
		static char readBuffer[readBufferSize];

		static bool connectedCOM;

		static char useSerial::inputBuffer[1024 * 1]; // store the input data here
		static bool useSerial::inputEntered;

	private:

		// Defines 
		static HANDLE hSerial;

		static std::string f_name;


		// Define the default values for serial connection parameters
		uint8_t dataBits = 8;  // Default data bits is 8
		uint8_t parity = NOPARITY;    // Default parity is 0 - accepts "NOPARITY"
		uint8_t stopBits = ONESTOPBIT;  // Default stop bits is 0 - accepts "ONESTOPBIT"
		int readTimeout = 5000;  // in ms, 5 seconds
		int writeTimeout = 5000; // in ms, 5 seconds


		
		
};

#endif // USE_SERIAL_H