#include "UseImGUI.h"
#include "SerialFunctions.h"
#include "windows.h"
#include "iostream"
#include "vector"
#include "string"
#include "Setupapi.h"
#include "devguid.h"
#include "cfgmgr32.h"

#pragma comment(lib, "setupapi.lib")


// Display Size Variables
const int displayWidth = 1280;
const int displayHeight = 720;


// Available Serial COM Ports List Variables
 static bool first_call = true;

static bool combo_is_open = false;
std::vector<std::string> com_ports_list; // Define a vector to store the available COM ports
static int selected_COMport = -1; // Define a variable to store the selected COM port
static int prev_selected_COMport = selected_COMport;


// Initializes the class variable that stores the Connected device's Friendly name
std::string useSerial::f_name;


// Initializes the class variable that stores state of program connection to any serial COM port
bool useSerial::connectedCOM = false;


// Initializes the class variable that stores the HANDLE of the current connection to COM port
HANDLE useSerial::hSerial;


// BAUD Rate Variables Variables
const int BAUD[] = { 300, 600, 750, 1200, 2400, 4800, 9600, 19200, 31250, 38400, 57600, 74880, 115200, 230400, 250000, 460800, 500000, 921600, 1000000, 2000000 };
static int selected_BAUD = 9600; // Here we store our selection data as an index. -1 means blank by default
static bool is_baud_combo_open = false; // Stores if the BAUD Combo List is open or closed


// Initializes an instance of the useSerial so it can be used with the readSerialData(buffer, sizeof(buffer)) && useSerial::connectedCOM command to read from serial
static useSerial serial;


// Variable that stores data read from serial.
char buffer[256] = { 0 };


// History Text Box Variables
const int messageLimit = 1024;
char Smonitor[messageLimit] = "Serial monitor loaded correctly. \n";
// static size_t smonitorLength = strlen(Smonitor);


// Serial Send End of Line Settings Combo List Varaibles
const char* EndOfLine[] = { "No Line Ending", "New Line", "Carriage Return", "Both NL & CR" }; // Define an array of items to display in the combo list
const char* EndOfLineChars[] = { "", "\r", "\n", "\r\n" };
int selected_EndOfLine = 0; // Define a variable to store the selected item index
static bool is_line_combo_open = false; // Stores if the BAUD Combo List is open or closed





// debug QueryDosDevice() com port list check
/*/ void printCOMPorts() {
	std::cout << "printCOMPorts: " << std::endl;
	char szDrive[4];
	char szDevices[65535];
	DWORD dwResult = QueryDosDevice(NULL, szDevices, 65535);

	if (dwResult == 0) {
		std::cerr << "Failed to retrieve list of devices: " << GetLastError() << std::endl;
		return;
	}

	std::cout << "Available COM ports: " << std::endl;
	char* pszDevice = szDevices;

	while (*pszDevice) {
		if (strncmp(pszDevice, "COM", 3) == 0) {
			std::cout << pszDevice << std::endl;
		}

		pszDevice += strlen(pszDevice) + 1;
	}
}/*/

// Get the COM Port name from a chosen index number
std::string getComPortName(int index, const std::vector<std::string>& com_ports_list) {
	if (index < 0 || index >= static_cast<int>(com_ports_list.size())) {
		return "";
	}
	else {
		// The COM port name is expected to be in the format "COMx", where x is a number
		std::string port_name = com_ports_list[index];
		size_t pos = port_name.find("COM");
		if (pos != std::string::npos) {
			port_name.erase(port_name.size() - 1); // remove the last character
			return port_name.substr(pos);
		}
		else {
			return "";
		}
	}
}


// Get an Integer Baud Rate value from a chosen index number




// Define a function to update the list of available COM ports
void updateAvailableSerialPortsComboList() {
	// std::cout << "updateAvailableSerialPortsComboList() is running \n" << std::endl;
	com_ports_list.clear(); // Clear the vector to start with a fresh list
	useSerial ports; // as the function isn't static, an instance of the class must be generated
	std::vector<std::string> ports_list = ports.getAvailableSerialPorts(); // the instance of the class is then used to call the function. For some reason...

	// Add each available COM port to the list
	for (const auto& port : ports_list) {
		// std::cout << "Available COM port For loop \n" << std::endl;
		com_ports_list.push_back(port);
	}

	// Reset the selected COM port index
	selected_COMport = -1;
	first_call = false;
	// std::cout << "Reset selected_index, end of updateAvailableSerialPortsComboList() \n" << std::endl;
}



void UseImGUI::Init(GLFWwindow* window, const char* glsl_version){
	IMGUI_CHECKVERSION();
	ImGui::CreateContext();
	ImGuiIO& io = ImGui::GetIO();
	ImGui_ImplGlfw_InitForOpenGL(window, true);
	ImGui_ImplOpenGL3_Init(glsl_version);
	ImGui::StyleColorsDark();
}

void UseImGUI::NewFrame(){
	ImGui_ImplOpenGL3_NewFrame();
	ImGui_ImplGlfw_NewFrame();
	ImGui::NewFrame();
}

void UseImGUI::Update(){

	ImGui::Begin("##Monitor", NULL, ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoCollapse | ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoSavedSettings);
	ImGui::SetWindowPos(ImVec2(0, 0));
	float winSize = ImGui::GetWindowSize().x;
	ImGui::SetWindowSize(ImVec2(displayWidth, displayHeight)); 	// 0.0f forces auto-fit on an axis to contents

	if (first_call) {
		Sleep(500);
		updateAvailableSerialPortsComboList();
	}

	if (ImGui::Button("Refresh List")) {
		std::cout << "Refresh button was pressed once \n" << std::endl;
		updateAvailableSerialPortsComboList();
		// printCOMPorts(); // Debug QueryDosDevice() function to check available serial ports with in a different way.
	}

	ImGui::SameLine();
	// Add a Dummy element with the set size to effectively space other elements
	ImGui::Dummy(ImVec2(5.0f, 0));
	ImGui::SameLine();

	ImGui::SetNextItemWidth(362);

	ImGui::Combo("COM Port", &selected_COMport,
		[](void* data, int idx, const char** out_text) {
			auto& com_ports = *static_cast<std::vector<std::string>*>(data);
			if (idx >= 0 && idx < static_cast<int>(com_ports.size())) {
				*out_text = com_ports[idx].c_str();
				if (!ImGui::IsItemHovered() && combo_is_open && prev_selected_COMport != selected_COMport) {
					std::cout << "COM Port" << selected_COMport << std::endl;
					combo_is_open = false;
				}
			}
			else {
				*out_text = "<Unknown>";
			}
			return true;
		}, static_cast<void*>(&com_ports_list), static_cast<int>(com_ports_list.size()), static_cast<int>(com_ports_list.size()) > 10);

	if (ImGui::IsItemHovered() && ImGui::IsItemActivated()) {
		combo_is_open = true;
		prev_selected_COMport = selected_COMport;
	}
	

	ImGui::SameLine();
	// Add a Dummy element with the set size to effectively space other elements
	ImGui::Dummy(ImVec2(20.0f, 0));
	ImGui::SameLine();

	ImGui::SetNextItemWidth(140); // force width


	if (ImGui::BeginCombo("BAUD Rate", std::to_string(selected_BAUD).c_str())) {
		is_baud_combo_open = true;
		for (int n = 0; n < IM_ARRAYSIZE(BAUD); n++) {
			const bool is_selected = (selected_BAUD == BAUD[n]);
			if (ImGui::Selectable(std::to_string(BAUD[n]).c_str(), is_selected))
				selected_BAUD = BAUD[n];

			if (is_selected) {
				ImGui::SetItemDefaultFocus();
			}
		}

		ImGui::EndCombo();
	}
	else {
		if (is_baud_combo_open) {
			// Combo list has been closed, do something here
			std::cout << "Baud: " << selected_BAUD << std::endl;
		}
		is_baud_combo_open = false;
	}

	ImGui::SameLine();
	// Add a Dummy element with the set size to effectively space other elements
	ImGui::Dummy(ImVec2(5.0f, 0));
	ImGui::SameLine();


	if (ImGui::Button("Connect")) {
		// std::cout << "Disconnect button clicked!" << std::endl;

		useSerial serialConnected;

		if (selected_COMport != -1) {
			
			// Get the COM port name based on the selected index
			std::string com_port_name = getComPortName(selected_COMport, com_ports_list);
			// std::cout << "UseImGUI:: Selected COM port: " << com_port_name << std::endl;

			// Print the selected baud rate
			// std::cout << "UseImGUI:: Selected baud rate: " << selected_BAUD << std::endl;

			// Create a useSerial object and attempt to connect to the serial port
			useSerial connectSerial;
			connectSerial.connectSerialPort(com_port_name, selected_BAUD);

			// std::cout << "UseImGUI:: Connected to serial?" << std::endl;
		}
		else {
			std::cout << "UseImGUI:: No COM port selected!" << std::endl;
		}
	}

	ImGui::SameLine();
	if (ImGui::Button("Disconnect")) {
		// std::cout << "Disconnect button clicked!" << std::endl;

		// Call the closeSerialPort function with the HANDLE variable
		useSerial closeSerial;
		closeSerial.closeSerialPort(selected_COMport);
	}


	// Temporarily modify the ImGuiCol_ChildBg color to match the default ImGui text box background color
	ImGui::PushStyleColor(ImGuiCol_ChildBg, ImVec4(ImGui::GetStyle().Colors[ImGuiCol_FrameBg]));

	// Begin the child window with the modified background color
	ImGui::BeginChild("##History", ImVec2(0, -ImGui::GetFrameHeightWithSpacing()), false, ImGuiWindowFlags_HorizontalScrollbar);

	// Get the ImDrawList for rendering
	ImDrawList* drawList = ImGui::GetWindowDrawList();

	// Get the position and size of the child window
	ImVec2 windowPos = ImGui::GetWindowPos();
	ImVec2 windowSize = ImGui::GetWindowSize();

	// Draw a filled rectangle with the modified background color
	drawList->AddRectFilled(windowPos, ImVec2(windowPos.x + windowSize.x, windowPos.y + windowSize.y), ImGui::GetColorU32(ImGui::GetStyle().Colors[ImGuiCol_ChildBg]));

	// Display the text inside the child window
	ImGui::TextWrapped("%s", Smonitor);

	// Scroll to the bottom if needed
	if (ImGui::GetScrollY() >= ImGui::GetScrollMaxY()) {
		ImGui::SetScrollHereY(1.0f);
	}

	// End the child window
	ImGui::EndChild();

	// Restore the original style color for ImGuiCol_ChildBg
	ImGui::PopStyleColor();


	if (ImGui::Button("Clear")) {
		Smonitor[0] = 0; // clear the buffer
		strcat_s(Smonitor, "History cleared\n"); // write default message
	}

	ImGui::SameLine();
	// Add a Dummy element with the set size to effectively space other elements
	ImGui::Dummy(ImVec2(10.0f, 0));
	ImGui::SameLine();
	



	static size_t smonitorLength = strlen(Smonitor);

	if (serial.readSerialData(buffer, sizeof(buffer)) && useSerial::connectedCOM) {
		buffer[useSerial::bytesRead] = '\0'; // make sure received data is null-terminated

		// check if adding the received data would exceed the buffer size
		if (smonitorLength + strlen(buffer) < sizeof(Smonitor)) {
			strcat_s(Smonitor, buffer); // append received data to Smonitor
			smonitorLength = strlen(Smonitor);
			memset(buffer, 0, sizeof(buffer)); // clear the buffer
		}
		else {
			// if adding the received data would exceed the buffer size,
			// trim the oldest message(s) and add the new data
			size_t trimmedLength = strlen(buffer);
			for (size_t i = 0; i < smonitorLength; i++) {
				if (Smonitor[i] == '\n') {
					trimmedLength += i + 1;
					smonitorLength -= i + 1;
					memmove(Smonitor, Smonitor + i + 1, smonitorLength + 1);
					i = 0;
				}
				if (smonitorLength + trimmedLength < sizeof(Smonitor)) {
					strcat_s(Smonitor, buffer); // append received data to Smonitor
					smonitorLength = strlen(Smonitor);
					memset(buffer, 0, sizeof(buffer)); // clear the buffer
					break;
				}
			}
		}
	}ImGui::SameLine();



	// Note: we are using a fixed-sized buffer for simplicity here. See ImGuiInputTextFlags_CallbackResize
	// and the code in misc/cpp/imgui_stdlib.h for how to setup InputText() for dynamically resizing strings.
	ImGui::SetNextItemWidth(800); // force width
	static char text[1024 * 1] = "";
	static bool textBoxClicked = false;
	if (ImGui::InputText("##KeyboardInput", text, IM_ARRAYSIZE(text), ImGuiInputTextFlags_EnterReturnsTrue)) {
		if (textBoxClicked) {

			std::cout << "Before concatenation: " << useSerial::inputBuffer << std::endl;
			strcpy_s(useSerial::inputBuffer, text); // copy the contents of the text box
			strcat_s(useSerial::inputBuffer, EndOfLineChars[selected_EndOfLine]);// Append the selected end-of-line character to the input buffer
			std::cout << "After concatenation: " << useSerial::inputBuffer << std::endl;
			text[0] = '\0'; // clear the contents of the text box
			useSerial::inputEntered = true; // set the flag to indicate that the input was entered

		}
	}

	if (ImGui::IsItemClicked()) {
		textBoxClicked = true;
	}
	else if (ImGui::IsMouseClicked(0)) {
		textBoxClicked = false;
	}

	if (ImGui::IsItemHovered()) {
		ImGui::SetTooltip("Press ENTER to send the text over serial");
	}

	ImGui::SameLine();

	// the ## means it won't display the label
	// The & on the buffer has to do with the data-type needed, a Char*. What ever that means


	ImGui::SetNextItemWidth(260); // force width

	if (ImGui::BeginCombo("##EndOfLine", EndOfLine[selected_EndOfLine])) {
		is_line_combo_open = true;
		for (int i = 0; i < 4; i++) {
			const bool is_selected = (selected_EndOfLine == i);
			if (ImGui::Selectable(EndOfLine[i], is_selected)) {
				selected_EndOfLine = i;
				std::cout << "Selected end of line: " << EndOfLine[selected_EndOfLine] << std::endl;
			}
			if (is_selected) {
				ImGui::SetItemDefaultFocus();
			}
		}
		ImGui::EndCombo();
	}
	else {
		if (is_line_combo_open) {
			// Combo list has been closed, do something here
			std::cout << "End Of Line" << selected_EndOfLine << std::endl;
		}
		is_line_combo_open = false;
	}


	ImGui::SameLine();

	if (ImGui::Button("Send", ImVec2(84, 0))) {
		std::cout << "Send button pressed" << std::endl;
		strcpy_s(useSerial::inputBuffer, text); // copy the contents of the text box
		text[0] = '\0'; // clear the contents of the text box
		useSerial::inputEntered = true; // set the flag to indicate that the input was entered
	}

	if (useSerial::inputEntered) {
		std::cout << "Sending input over serial" << std::endl;
		serial.sendSerialData(useSerial::inputBuffer, strlen(useSerial::inputBuffer)); // send the input data over serial
		useSerial::inputEntered = false; // reset the flag
	}
	ImGui::End();
}

void UseImGUI::Render() {
	ImGui::Render();
	ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());
}

void UseImGUI::Shutdown() {
	ImGui_ImplGlfw_Shutdown();
	ImGui_ImplOpenGL3_Shutdown();
	ImGui::DestroyContext();
}



